<?php

namespace App\Console\Commands;

use Illuminate\Console\Command;
use Illuminate\Support\Facades\DB;

class GenerateTreeCommand extends Command
{

    private $routesFile = null;

    /**
     * The name and signature of the console command.
     *
     * @var string
     */
    protected $signature = 'generatetree
                            {path: /path/to/tree.txt}';

    /**
     * Execute the console command.
     *
     * @return mixed
     */
    public function handle()
    {

        $lines = ['<?php', '', 'use Illuminate\Support\Facades\Route;', '', ''];
        file_put_contents(base_path("routes/generated.php"), implode("\n", $lines));

        $sideLinks = [];

        global $argv;
        $file = fopen($argv[2], "r");
        $lines = [];
        while (!feof($file)) {
            $line = rtrim(fgets($file));
            if(trim($line) !== '') {
                $lines[] = $line;
            }
        }

        fclose($file);

        $currentRoot = "";
        $currentController = new GenController();
        $currentSubController = new GenController();
        $currentSubType = "";
        $currentView = "";
        $currentMethod = null;

        foreach ($lines as $line) {

            // skip comments
            if(trim($line)[0] === '#') continue;

            $lineType = null;

            // no leading space - root specifier
            if($line[0] !== ' ') {
                $currentRoot = trim($line);
            }
            else {

                $ls = $this->numLS($line);
                $line = trim($line);

                $tokens = explode("|", $line);
                $line = $tokens[0];
                $dbTable = null;
                if(count($tokens) >= 2) {
                    $dbTable = $tokens[1];
                }

                $hasAdd = in_array("add", $tokens);
                $hasView = in_array("view", $tokens);

                switch($ls) {

                    case 4: // top level controller OR top level controller action

                        // top level controller-action
                        if(strpos($line, "/") !== FALSE) {
                            if(!empty($currentController)) {
                                $method = explode("/", $line)[1];
                                $newMethod = $currentController->addMethod($method, "/" . $line);
                                // create _SINGLE_ controller if view
                                if($method === "view") {
                                    $currentSubController = new GenController($currentRoot, $currentController->name . "_SINGLE");
                                    $currentSubController->dbTable = $currentController->dbTable;
                                    $currentSubController->parentRoute = "/" . $line;
                                    $currentSubController->parentControllerName = $currentController->name;
                                    $currentSubController->sub = true;
                                    $newMethod->redirect = "/" . $line . "/SUB_dashboard";
                                }
                            }
                        }
                        // new top level controller
                        else if(empty($currentController) || $line !== $currentController->name) {
                            if(!empty($currentController)) {
                                $currentController->save();
                            }
                            $currentController = new GenController($currentRoot, $line);
                            $currentController->dbTable = $dbTable;
                            $currentController->hasAdd = $hasAdd;
                            $currentController->hasView = $hasView;
                            $currentController->addMethod("index", "/$line");
                            if(!empty($currentSubController)) {
                                $currentSubController->save();
                            }
                            $currentSubType = '';
                            $currentSubController = new GenController();

                            $sideLinks[] = "<li class='nav-item'><a href='/{$currentController->name}' " .
                                "class='nav-link " .
                                "{{ (isset(request()->route()->getController()->selfName) && strpos(request()->route()->getController()->selfName, '{$currentController->name}') === 0 ? 'active' : '') }}" . " '>" .
                                "<i class='nav-icon fa fa-user'></i>" .
                                "<p>" . $currentController->snakeToTitleCase($currentController->name) . "</p>" .
                                "</a></li>";
                        }
                        break;

                    case 8: // sub-type declaration

                        $currentSubType = $line;
                        break;

                    case 12: // subs and actions

                        if($currentSubType === 'ACTIONS') {
                            $currentMethod = $currentSubController->addMethod(
                                "ACTION_" . $line,
                                "/ACTION_" . $line
                            );
                            $currentMethod->type = 'action';
                            $currentMethod->data = [];
                        }
                        else if($currentSubType === 'SUB') {
                            $currentMethod = $currentSubController->addMethod(
                                "SUB_" . $line,
                                "/SUB_" . $line
                            );
                            $currentMethod->type = 'sub';
                            $currentMethod->data = [];
                        }

                        break;

                    case 16: // data for actions and subs

                        if(!empty($currentMethod)) {
                            $currentMethod->data[] = $line;
                        }

                        break;

                }

            }
        }

        // do any pending saves
        if(!empty($currentSubController)) {
            $currentSubController->save();
        }
        if(!empty($currentController)) {
            $currentController->save();
        }

        echo "Saved " . base_path("routes/generated.php") . "\n";

        // save side links
        file_put_contents(resource_path("views/layouts/generated-links.blade.php"), implode("\n", $sideLinks));
        echo "Saved " . resource_path("views/layouts/generated-links.blade.php") . "\n";

    }

    private function numLS($line) {
        $count = 0;
        for ($i=0; $i<strlen($line); $i++) {
            if($line[$i] !== ' ') break;
            $count++;
        }
        return $count;
    }

}

class GenController {
    public $root;
    public $saved = false;
    public $name;
    public $methods;
    public $parentRoute = "";
    public $dbTable = null;
    public $hasAdd = false;
    public $hasView = false;
    public $sub = false;
    public $parentControllerName = '';
    public $subLinksSaved = false;
    public $actionLinksSaved = false;
    public function __construct($root = null, $name = null)
    {
        $this->root = $root;
        $this->name = $name;
        $this->methods = [];
    }
    public function addMethod($method, $route) {
        if($this->parentRoute) {
            $route = $this->parentRoute . $route;
        }
        $method = new GenControllerMethod($method, $route);
        $this->methods[] = $method;
        return $method;
    }
    public function save() {
        if(!$this->saved && !empty($this->root) && !empty($this->name)) {
            $this->saveController();
            $this->saveRoutes();
            $this->saved = true;
            $this->log();
        }
    }
    public function saveController() {
        $text = file_get_contents(base_path('generatecv/tree-templates/controller.template.php'));
        $code = [];
        foreach ($this->methods as $method) {
            $code[] = "";
            $code[] = "\t// GET {$method->route}";
            $code[] = "\t" . 'public function ' . $method->name . '(Request $request' . ($method->hasUID ? ', $uid' : '') . ') {';

            if($method->redirect) {
                $target = str_replace('{uid}', '$uid', $method->redirect);
                $code[] = "\t\t" . 'return redirect("' . $target . '");';
            }
            else {
                if($method->hasUID) {
                    $code[] = "\t\t\$record = DB::table('{$this->dbTable}')->where('uid', \$uid)->first();";
                    $code[] = "\t\treturn view('{$this->root}/{$this->name}/{$method->name}', compact('record'));";
                }
                else {
                    $code[] = "\t\t\$records = DB::table('{$this->dbTable}')->get();";
                    $code[] = "\t\treturn view('{$this->root}/{$this->name}/{$method->name}', compact('records'));";
                }
            }

            $this->saveView($this, $method);

            $code[] = "\t}";
        }
        $text = str_replace("_NAME_", "{$this->name}_Controller", $text);
        $text = str_replace("// __METHODS__", implode("\n", $code), $text);
        file_put_contents(app_path("Http/Controllers/{$this->name}_Controller.php"), $text);
        echo "Generated " . app_path("Http/Controllers/{$this->name}_Controller.php") . "\n";
    }
    public function saveView(GenController $controller, GenControllerMethod $method) {

        if($controller->sub) {
            $controller->saveSubLinks($controller, $method);
            $controller->saveActionLinks($controller, $method);
            if($method->type === 'action') {
                $this->saveSubActionView($controller, $method);
            }
            else if($method->type === 'sub' && count($method->data)) {
                $this->saveSubSubView($controller, $method);
            }
            else {
                $this->saveSubDefaultView($controller, $method);
            }
        }
        else {
            if($method->name === 'view' && $controller->hasView) {
                $this->saveShowView($controller, $method);
            }
            else if($method->name === 'index') {
                $this->saveIndexView($controller, $method);
            }
            else if($method->name === 'add_new' && $controller->hasAdd) {
                $this->saveAddNewView($controller, $method);
            }
        }

    }
    public function saveIndexView(GenController $controller, GenControllerMethod $method)
    {
        $text = file_get_contents(base_path('generatecv/tree-templates/index.template.blade.php'));
        $text = str_replace("_NAME_", $this->snakeToTitleCase($controller->name), $text);

        if($controller->hasAdd) {
            $addLink = "<a class='btn btn-primary btn-sm' href='/{$controller->name}/add_new'>" .
                "<i class='fa fa-plus-circle' aria-hidden='true'></i> Add New</a>";
            $text = str_replace("<!-- _ADD_NEW_LINK_ -->", $addLink, $text);
        }

        $columns = DB::getSchemaBuilder()->getColumnListing($controller->dbTable);
        $ths = [];
        $tds = [];
        foreach ($columns as $column) {
            $ths[] = "<th>{$this->snakeToTitleCase($column)}</th>";
            $tds[] = "<td>" .
                ($controller->hasView && $column === 'uid' ? '<a href="/' . $controller->name . '/view/<?= $record->uid ?>">' : '') .
                "<?= \$record->$column ?>" .
                ($controller->hasView && $column === 'uid' ? '</a>' : '') .
                "</td>";
        }
        $text = str_replace("<!-- __SCAFFOLD_THS__ -->", implode("\n", $ths), $text);
        $text = str_replace("<!-- __SCAFFOLD_TDS__ -->", implode("\n", $tds), $text);
        $this->file_force_contents(resource_path("views/{$controller->root}/{$controller->name}/{$method->name}.blade.php"), $text);
        echo "Generated " . resource_path("views/{$controller->root}/{$controller->name}/{$method->name}.blade.php") . "\n";
    }
    public function saveShowView(GenController $controller, GenControllerMethod $method)
    {
        // delete sub links and action links
        if(file_exists(resource_path("views/{$controller->root}/{$controller->parentControllerName}/subs.blade.php"))) {
            unlink(resource_path("views/{$controller->root}/{$controller->parentControllerName}/subs.blade.php"));
        }
        if(file_exists(resource_path("views/{$controller->root}/{$controller->parentControllerName}/actions.blade.php"))) {
            unlink(resource_path("views/{$controller->root}/{$controller->parentControllerName}/actions.blade.php"));
        }

        $text = file_get_contents(base_path('generatecv/tree-templates/show.template.blade.php'));
        $text = str_replace("_NAME_", $this->snakeToTitleCase($controller->name), $text);
        $text = str_replace("_UID_", '<?= $record->uid ?>', $text);
        $text = str_replace("_INDEX_ROUTE_", $controller->name . '-index', $text);
        $text = str_replace("_SUB_LINKS_VIEW_", "{$controller->root}/{$controller->name}/subs", $text);
        $this->file_force_contents(resource_path("views/{$controller->root}/{$controller->name}/{$method->name}.blade.php"), $text);
        echo "Generated " . resource_path("views/{$controller->root}/{$controller->name}/{$method->name}.blade.php") . "\n";
    }
    public function saveSubLinks(GenController $controller, GenControllerMethod $method)
    {
        if ($controller->subLinksSaved) return;
        $subLinksView = resource_path("views/{$controller->root}/{$controller->parentControllerName}/subs.blade.php");
        $subLinks = [];
        foreach ($controller->methods as $meth) {
            if (strpos($meth->name, "SUB_") !== 0) continue;
            $display = $this->snakeToTitleCase(substr($meth->name, 4));
            $subLinks[] = "<a " .
                "href='/{$controller->parentControllerName}/view/<?= \$record->uid ?>/{$meth->name}' " .
                "class='d-block p-3 py-2 border-bottom " .
                "{{ request()->route()->getActionMethod() === '{$meth->name}' ? 'bg-secondary text-white' : '' }}" .
                ($meth->name === 'SUB_dashboard' ? "{{ strpos(request()->route()->getActionMethod(), 'ACTION_') === 0 ? 'bg-secondary text-white' : '' }}" : "")
                . "'>$display</a>";
        }
        file_put_contents($subLinksView, implode("\n", $subLinks));
        echo "Generated " . $subLinksView . "\n";
        $controller->subLinksSaved = true;
    }
    public function saveActionLinks(GenController $controller, GenControllerMethod $method)
    {
        if ($controller->actionLinksSaved) return;
        $actionLinksView = resource_path("views/{$controller->root}/{$controller->parentControllerName}/actions.blade.php");
        $actionLinks = [];
        foreach ($controller->methods as $meth) {
            if (strpos($meth->name, "ACTION_") !== 0) continue;
            $display = $this->camelToTitleCase(substr($meth->name, 7));
            $actionLinks[] = "<a " .
                "href='/{$controller->parentControllerName}/view/<?= \$record->uid ?>/{$meth->name}' " .
                "class='d-block btn btn-sm btn-default mb-3'>$display</a>";
        }
        file_put_contents($actionLinksView, implode("\n", $actionLinks));
        echo "Generated " . $actionLinksView . "\n";
        $controller->actionLinksSaved = true;
    }
    public function saveSubDefaultView(GenController $controller, GenControllerMethod $method)
    {
        $text = file_get_contents(base_path('generatecv/tree-templates/sub.template.blade.php'));
        $text = str_replace("_LAYOUT_", "{$controller->root}.{$controller->parentControllerName}.view", $text);
        $text = $this->generateSubContent($controller, $method, $text);
        $this->file_force_contents(resource_path("views/{$controller->root}/{$controller->name}/{$method->name}.blade.php"), $text);
        echo "Generated " . resource_path("views/{$controller->root}/{$controller->name}/{$method->name}.blade.php") . "\n";
    }
    public function saveSubActionView(GenController $controller, GenControllerMethod $method) {
        $text = file_get_contents(base_path('generatecv/tree-templates/sub-action.template.blade.php'));
        $text = str_replace("_LAYOUT_", "{$controller->root}.{$controller->parentControllerName}.view", $text);
        $text = str_replace("_NAME_", $this->camelToTitleCase($this->snakeToTitleCase($method->name)), $text);
        $text = str_replace("_BACK_ROUTE_", "{$controller->parentControllerName}-view", $text);
        $text = str_replace("_API_", "/api/{$controller->dbTable}/" . substr($method->name, 7), $text);
        $text = str_replace("_RETURN_ROUTE_", "{$controller->name}-{$method->name}", $text);
        $fields = [];
        if(count($method->data)) {
            foreach ($method->data as $field) {
                $fields[] =
                    "<div class='form-group mb-3'>" .
                    "<label class='control-label'>{$this->camelToTitleCase($this->snakeToTitleCase($field))}</label>" .
                    "<input class='form-control' type='text' name='$field'>" .
                    "</div>";
            }
        }
        $text = str_replace("<!-- _SCAFFOLD_FIELDS_ -->", implode("\n", $fields), $text);
        $this->file_force_contents(resource_path("views/{$controller->root}/{$controller->name}/{$method->name}.blade.php"), $text);
        echo "Generated " . resource_path("views/{$controller->root}/{$controller->name}/{$method->name}.blade.php") . "\n";
    }
    public function saveSubSubView(GenController $controller, GenControllerMethod $method) {

    }
    public function generateSubContent(GenController $controller, GenControllerMethod $method, $text) {
        if($method->name === 'SUB_dashboard') {
            $html = file_get_contents(base_path('generatecv/tree-templates/dashboard.template.blade.php'));
            $text = str_replace("_SUB_VIEW_", $html, $text);
            $text = str_replace("_ACTION_LINKS_VIEW_", "{$controller->root}/{$controller->parentControllerName}/actions", $text);
        }
        else {
            $text = str_replace("_SUB_VIEW_",
                "<h4 class='py-3 border-bottom'>" .
                $this->camelToTitleCase($this->snakeToTitleCase($method->name)) . "</h4>" .
                "Controller: <b>{$controller->name}</b><br>" .
                "Action: <b>{$method->name}()</b><br>" .
                "View: <b>{$controller->root}/{$controller->name}/{$method->name}.blade.php</b><br>",
                $text);
        }
        return $text;
    }
    public function saveAddNewView(GenController $controller, GenControllerMethod $method)
    {
        $text = file_get_contents(base_path('generatecv/tree-templates/add_new.template.blade.php'));
        $text = str_replace("_NAME_", $this->snakeToTitleCase($controller->name), $text);
        $text = str_replace("_BACK_ROUTE_", "{$controller->name}-index", $text);
        $columns = DB::getSchemaBuilder()->getColumnListing($controller->dbTable);
        $fields = [];
        foreach ($columns as $column) {
            if(in_array($column,
                    ['id', 'uid', 'created_at',
                        'deactivated_at', 'deactivation_memo',
                        'reactivated_at', 'reactivation_memo']) === true) continue;
            $fields[] =
                "<div class='form-group mb-3'>" .
                    "<label class='control-label'>{$this->snakeToTitleCase($column)}</label>" .
                    "<input class='form-control' type='text' name='{$this->snakeToCamelCase($column)}'>" .
                "</div>";
        }
        $text = str_replace("<!-- _SCAFFOLD_FIELDS_ -->", implode("\n", $fields), $text);
        $this->file_force_contents(resource_path("views/{$controller->root}/{$controller->name}/{$method->name}.blade.php"), $text);
        echo "Generated " . resource_path("views/{$controller->root}/{$controller->name}/{$method->name}.blade.php") . "\n";
    }
    public function saveRoutes() {
        $lines = ["// --- {$this->root}: {$this->name} --- //"];
        foreach ($this->methods as $method) {
            // FORMAT:
            // Route::get('/foo/bar/{uid}', 'FooController@bar')->name('foo-action');
            $lines[] = "Route::get('{$method->route}', '{$this->name}_Controller@{$method->name}')->name('{$this->name}-{$method->name}');";
        }
        $lines[] = '';
        $lines[] = '';
        file_put_contents(base_path("routes/generated.php"), implode("\n", $lines), FILE_APPEND);
    }
    public function log() {
        $this->w('');
        $this->w("Controller: app/Http/Controllers/{$this->name}_Controller");
        $this->w("Table: {$this->dbTable}");
        $this->w('---------------------------------------------');
        foreach ($this->methods as $method) {
            $this->w('Rout: ' . $method->route, 1);
            $this->w('Meth: ' . $method->name . '($request' . ($method->hasUID ? ', $uid' : '') . ')', 1);
            if(!empty($method->data)) $this->w('Data: ' . implode(", ", $method->data), 1);
            if(!$method->redirect) {
                $this->w('View: ' . resource_path("views/{$this->root}/{$this->name}/{$method->name}.blade.php"), 1);
            }
            else {
                $this->w('Redi: ' . $method->redirect, 1);
            }
            $this->w('---------------------------------------------');
        }
    }
    public function w($line, $level = -1) {
        for($i=0; $i<$level; $i++) echo "\t";
        echo "$line\n";
    }
    public function snakeToTitleCase($text) {
        $text = preg_replace("/^(SUB|ACTION)_/", "", $text);
        return ucwords(str_replace("_", " ", $text));
    }
    public function camelToTitleCase($text) {
        $text = preg_replace("/^(SUB|ACTION)_/", "", $text);
        $text = preg_replace("/([a-z])([A-Z0-9])/", "$1 $2", $text);
        return ucwords($text);
    }
    public function snakeToCamelCase($text) {
        $text = ucwords(str_replace("_", " ", $text));
        $text = str_replace(" ", "", $text);
        $text[0] = strtolower($text[0]);
        return $text;
    }
    private function file_force_contents($dir, $contents){
        $dir = str_replace("\\", "/", $dir);
        $dir = str_replace( '//', '/', $dir);
        $parts = explode('/', $dir);
        $file = array_pop($parts);
        $dir = '';
        foreach($parts as $part) {
            if($part[strlen($part) - 1] !== ':') {
                if(!is_dir($dir .= "/$part")) mkdir($dir);
            }
        }
        file_put_contents("$dir/$file", $contents);
    }
}

class GenControllerMethod {
    public $name;
    public $route;
    public $hasUID = false;
    public $redirect = false;
    public $type = '';
    public $data = [];
    public function __construct($name, $route)
    {
        $this->name = $name;
        $this->route = $route;
        if(strpos($this->route, "{uid}") !== FALSE) {
            $this->hasUID = true;
        }
    }
}
